unit URechner;

(*  Einige Anmerkungen zum Programm und zu der Aufgabe :
    ====================================================
    Die eingegebenen Zahlen und die mathematischen Verknpfungen werden in eine
    einfach verkettete Liste eingeordnet, wobei das Element 'n' die n-te Zahl
    und die Verknpfung mit der (n+1)-ten Zahl enthlt. Ein Ergebnis wird mit
    der '='-Taste angefordert. Gleichzeitig wird der Listenkopf freigegeben. Das
    Ergebnis kann Ausgangspunkt fr eine neue Rechenkette sein.

    Funktionswerte werden sofort berechnet und in 'n' gespeichert. Reihenfolge:
    z.B. sin -> 1.25 -> = (wie bei den neueren Rechnern).

    Wegen der einfach verketteten Liste arbeitet - ungewohnt - das Programm
    den Rechenterm von rechts nach links ab. Dabei werden alle hheren Rechen-
    arten nach Rangfolge schlielich auf eine arithmetische Summe zurckgefhrt,
    die dann berechnet wird (ohne Listenelemente per Zeigerverbiegen zu elimi-
    nieren). Die Routine wird dadurch zwar lnger, ist aber von Schlern so bes-
    ser zu durchschauen als der elegante Weg : doppelt verkettete Liste, da man
    von links nach rechts rechnet, Potenz- vor Punkt- vor Strichrechnung, wobei
    mit dem Vereinfachen von Termen Zeiger verbogen werden, bis schlielich nur
    eine arithmetische Summe brigbleibt.

    Wenn Klammerrechnung zustzlich ber Tasten eingebaut werden soll, mu man
    rekursiv arbeiten (Prozeduren TForm1.gleichGeklickt/TForm1.zahlEingabe
    rekursiv aufrufen).

    Einige Besonderheiten: die x-hoch-y-Taste arbeitet mit der ln-Funktion, kann
    also nichts mit negativen Basen anfangen (wie die Taschenrechner, wenn der
    Exponent keine ganze Zahl ist). Fr ganzzahlige Exponenten bietet Delphi die
    Funktion  'IntPower', die also auch Terme wie -3.123 hoch -7 berechnet.
    Zur Umsetzung hatte ich aber keine Lust mehr ...also Exponenten auf Ganz-
    zahligkeit prfen und alternativ zu ln(x) die Funktion 'IntPower' aufrufen.

    Sollte sich die Rechnung irgendwo aufhngen, so gibt es die Taste ClearAll..

    Die Behandlung von negativen Radikanden bei der Sqrt-Funktion durch Delphi
    habe ich nicht verstanden. Gibt man sqrt(-25) ein, so erhlt man ca. '0',
    also 4 E -305.   ??????   Verstehe ich nicht.
                                                          Tiemann
                                 *)


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, mKnopf, mTextfeld, Math;
                                                    //Unit Math hinzugefgt
type
  TForm1 = class(TForm)
    EingabeFeld               : TTextfeld;
    ErrorFeld                 : TTextfeld;
    einsKnopf                 : TKnopf;
    fuenfKnopf                : TKnopf;
    sechsKnopf                : TKnopf;
    zweiKnopf                 : TKnopf;
    vierKnopf                 : TKnopf;
    dreiKnopf                 : TKnopf;
    siebenKnopf               : TKnopf;
    durchKnopf                : TKnopf;
    neunKnopf                 : TKnopf;
    vorzeichenKnopf           : TKnopf;
    achtKnopf                 : TKnopf;
    nullKnopf                 : TKnopf;
    dezimalKnopf              : TKnopf;
    minusKnopf                : TKnopf;
    plusKnopf                 : TKnopf;
    malKnopf                  : TKnopf;
    gleichKnopf               : TKnopf;
    clearKnopf                : TKnopf;
    prozKnopf                 : TKnopf;
    rueckKnopf                : TKnopf;
    storeKnopf                : TKnopf;
    piKnopf                   : TKnopf;
    expKnopf                  : TKnopf;
    wurzelKnopf               : TKnopf;
    tanKnopf                  : TKnopf;
    cosKnopf                  : TKnopf;
    sinKnopf                  : TKnopf;
    clearLetztesZeichenKnopf  : TKnopf;
    clearEingabeKnopf         : TKnopf;
    promiKnopf                : TKnopf;


    procedure clearEingabeFeld(Sender: TObject);
    procedure einsGeklickt(Sender: TObject);
    procedure zweiGeklickt(Sender: TObject);
    procedure dreiGeklickt(Sender: TObject);
    procedure vierGeklickt(Sender: TObject);
    procedure fuenfGeklickt(Sender: TObject);
    procedure sechsGeklickt(Sender: TObject);
    procedure siebenGeklickt(Sender: TObject);
    procedure achtGeklickt(Sender: TObject);
    procedure neunGeklickt(Sender: TObject);
    procedure nullGeklickt(Sender: TObject);
    procedure dezimalGeklickt(Sender: TObject);
    procedure vorzeichenGeklickt(Sender: TObject);
    procedure clearEingabeGeklickt(Sender: TObject);
    procedure letztesZeichenGeklickt(Sender: TObject);
    procedure piGeklickt(Sender: TObject);
    procedure stoGeklickt(Sender: TObject);
    procedure rclGeklickt(Sender: TObject);
    procedure clearAllGeklickt(Sender: TObject);
    procedure plusGeklickt(Sender: TObject);
    procedure minusGeklickt(Sender: TObject);
    procedure malGeklickt(Sender: TObject);
    procedure durchGeklickt(Sender: TObject);
    procedure gleichGeklickt(Sender: TObject);
    procedure sinusGeklickt(Sender: TObject);
    procedure cosGeklickt(Sender: TObject);
    procedure tanGeklickt(Sender: TObject);
    procedure wurzelGeklickt(Sender: TObject);
    procedure expGeklickt(Sender: TObject);
    procedure prozGeklickt(Sender: TObject);
    procedure promiGeklickt(Sender: TObject);


  private
    procedure zahlEingabe(ziffer:Char);       //Verwaltung der Zifferneingabe
    procedure verketten (symbol: string);     //Operatorkette in Liste einordnen
    function funktional(argument:string):real;//Bearbeitung v. Funktionseingaben
    procedure errorMeldung;                   //Division durch Null; ln(-5) etc.
    { Private-Deklarationen }

  public
    { Public-Deklarationen }
  end;

  
  type kette  = ^zahlenKette;                 //Operatorkette
  zahlenkette = record
                 zahlen     : Real;           //Zahlenwerte
                 knuepfen   : string[5];      //Verknpfungen
                 vorgaenger : kette
                end;

  var

  v, z                 : kette;               //Hilfsvariablen
  letzterWert          : kette ;              //Listenkopf
  Form1                : TForm1;
  eingabeZahl          : String;              //Zahl im Textfeld
  speicher             : Real = 0.0;          //numerisch: STO-Taste
  funktionsArt         : char;                //sin als 's', cos als 'c' etc.

  piGeladen            : Boolean =False;      //fr Taste Pi
  gespeichert          : Boolean = false;     //fr Taste RCL
  dezimalPunkt         : Boolean = False;     //Integer o. Real im Eingabefeld ?
  dezimalPunktSpeicher : Boolean = False;     // s.o.
  verknuepfung         : Boolean = False;     //war letzte Eingabe Verknpfung ?
  bildetKette          : Boolean = false;     //ist Operatorkette angelegt ?
  funktionGeklickt     : Boolean = false;     //fr Funktionstasten
  gleichWarGeklickt    : Boolean = false;     //fr Gleich-Taste

implementation

{$R *.DFM}


//-------------- Textfelder --------------------------------
procedure TForm1.clearEingabeFeld(Sender: TObject);
begin
  eingabeFeld.clear;
  dezimalPunkt:=false;
end;
//----------------------------------------------------------
procedure TForm1.errorMeldung;
begin
  errorFeld._istSichtbar:=True;
end;
//----------------------------------------------------------
//-------------- Tasten ------------------------------------
//----------------------------------------------------------
procedure TForm1.einsGeklickt(Sender: TObject);
begin
 zahlEingabe(einsKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.zweiGeklickt(Sender: TObject);
begin
  zahlEingabe(zweiKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.dreiGeklickt(Sender: TObject);
begin
  zahlEingabe(dreiKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.vierGeklickt(Sender: TObject);
begin
  zahlEingabe(vierKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.fuenfGeklickt(Sender: TObject);
begin
  zahlEingabe(fuenfKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.sechsGeklickt(Sender: TObject);
begin
  zahlEingabe(sechsKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.siebenGeklickt(Sender: TObject);
begin
  zahlEingabe(siebenKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.achtGeklickt(Sender: TObject);
begin
  zahlEingabe(achtKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.neunGeklickt(Sender: TObject);
begin
  zahlEingabe(neunKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.nullGeklickt(Sender: TObject);
begin
  zahlEingabe(nullKnopf._Inhalt[2]);
end;
//----------------------------------------------------------
procedure TForm1.dezimalGeklickt(Sender: TObject);
begin
  if not dezimalPunkt then
  zahlEingabe(dezimalKnopf._Inhalt[2]);
  dezimalPunkt:= true;                  //pro Zahl nur ein Dezimalpunkt mglich
end;

//----------------------------------------------------------
//-------------Verwaltung der Zahlen-Eingabe ---------------
//----------------------------------------------------------

procedure TForm1.zahlEingabe(ziffer:Char);
begin
  errorFeld._istSichtbar:=false;
  If length(eingabeZahl)=0 Then eingabeZahl:='+';  //Pluszeichen vor jede Zahl
  If (length(eingabeZahl) < 12) then
    if  not(piGeladen) and not(gespeichert)  then  //Pi wird nicht verndert
         Begin
           eingabeZahl:=eingabeZahl+ ziffer;
           eingabeFeld._inhalt:=eingabeZahl;
           verknuepfung:=false;                    //erst im Anschlu an Zahlen-
         end                                       //eingabe ist Verknpfung
       else                                        //wieder sinnvoll
         begin
            self.clearEingabeGeklickt(self);
            setLength(eingabeZahl,0);
            eingabeZahl:='+';
            eingabeZahl:=eingabeZahl+ ziffer;
            eingabeFeld._inhalt:=eingabeZahl;
            piGeladen :=false;
            gespeichert:=false;
            dezimalPunkt:=false;
            verknuepfung:=false;                   //Verknpfung ist wieder
         end                                       //   sinnvoll
end;
//----------------------------------------------------------
procedure TForm1.vorzeichenGeklickt(Sender: TObject);
begin
 If length(eingabeZahl) >0 Then                    //Umschaltung +/-
  If eingabeZahl[1]='+' then eingabeZahl[1]:='-'
                        else eingabeZahl[1]:='+';
 eingabeFeld._inhalt:=eingabeZahl;
end;
//----------------------------------------------------------
procedure TForm1.clearEingabeGeklickt(Sender: TObject);
begin
  errorFeld._istSichtbar:=false;
  eingabeFeld.clear;
  setLength(eingabeZahl,0);
  dezimalPunkt:=false;
end;
//----------------------------------------------------------
procedure TForm1.letztesZeichenGeklickt(Sender: TObject);
Var  long : Integer;
begin
  long:= length(eingabeZahl);
  If long > 0 then                         //Lschen einzelner Zeichen zur
         begin                             //Korrektur
           if eingabeZahl[long]='.' then
                 dezimalPunkt:=false;
           setLength(eingabeZahl,long-1);
         end;
  eingabeFeld._inhalt:=eingabeZahl;
end;
//----------------------------------------------------------
procedure TForm1.piGeklickt(Sender: TObject);
begin
  self.clearEingabeGeklickt(self);          // Pi eingeben
  eingabeZahl:='+3.141592654';
  eingabeFeld._inhalt:=eingabeZahl;
  piGeladen:=True;
  verknuepfung:=false;
end;
//----------------------------------------------------------
procedure TForm1.stoGeklickt(Sender: TObject);
begin
  speicher:=eingabeFeld.inhaltAlsZahl;      //Speicher belegen: STO-Taste
  dezimalPunktSpeicher:=dezimalPunkt;
end;
//----------------------------------------------------------
procedure TForm1.rclGeklickt(Sender: TObject);
begin
  self.clearEingabeGeklickt(self);
  if speicher < 0 then eingabeFeld._inhalt:=floatToStr(speicher)
                  else eingabeFeld._inhalt:='+'+floatToStr(speicher);
  eingabeZahl := eingabeFeld._inhalt;
  dezimalPunkt:=dezimalPunktSpeicher;       //Speicher benutzen : RCL-Taste
  gespeichert:=true;
  verknuepfung:=false;
end;
//----------------------------------------------------------
procedure TForm1.clearAllGeklickt(Sender: TObject);
begin
   errorFeld._istSichtbar:=false;
   self.clearEingabeGeklickt(self);

   if bildetKette then dispose(letzterWert);

   piGeladen:=false;
   dezimalPunkt:=false;
   dezimalPunktSpeicher:=false;
   bildetKette:=false;
   speicher:=0.0;
   gespeichert:=false;
end;

//==========================================================
//================== Berechnungen ==========================
//==========================================================


procedure TForm1.plusGeklickt(Sender: TObject);
begin
  verketten('+');
  self.clearEingabeGeklickt(self);
end;
 //----------------------------------------------------------
procedure TForm1.minusGeklickt(Sender: TObject);
begin
  verketten('-');
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.malGeklickt(Sender: TObject);
begin
  verketten('*');
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.durchGeklickt(Sender: TObject);
begin
  verketten('/');
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.expGeklickt(Sender: TObject);
begin
  verketten('e');
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.prozGeklickt(Sender: TObject);
begin
  verketten('p');                   //Prozenttaste: '20 Taste 80 =' ---> '25'
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.promiGeklickt(Sender: TObject);
begin
  verketten('m');                   //Promilletaste: '0.2 Taste 75 =' --> '2.66'
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.sinusGeklickt(Sender: TObject);
begin
  funktionGeklickt:=true;
  verknuepfung:=false;
  gleichWarGeklickt:=false;
  funktionsArt:='s';
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.cosGeklickt(Sender: TObject);
begin
  funktionGeklickt:=true;
  verknuepfung:=false;
  gleichWarGeklickt:=false;
  funktionsArt:='c';
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.tanGeklickt(Sender: TObject);
begin
  funktionGeklickt:=true;
  verknuepfung:=false;
  gleichWarGeklickt:=false;
  funktionsArt:='t';
  self.clearEingabeGeklickt(self);
end;
//----------------------------------------------------------
procedure TForm1.wurzelGeklickt(Sender: TObject);
begin
  funktionGeklickt:=true;
  verknuepfung:=false;
  gleichWarGeklickt:=false;
  funktionsArt:='w';
  self.clearEingabeGeklickt(self);
end;

//----------------------------------------------------------
//---------------Operatorkette aufbauen---------------------
//----------------------------------------------------------

procedure TForm1.verketten (symbol : string);
begin
  gleichWargeklickt:=false;
  if not bildetkette then
       begin
         if length(eingabeZahl)=0 then exit;  //Verknpfung ohne vorhandene Zahl
         new(letzterWert);
         letzterWert.vorgaenger:=nil;
         if funktionGeklickt then letzterWert.zahlen:=funktional(eingabeZahl)
                             else letzterWert.zahlen:=strToFloat(eingabeZahl);
         letzterWert.knuepfen:=symbol;
         bildetKette:=true;
       end
       else
         if  not verknuepfung  then
             begin
               new(z);
               z.vorgaenger:=letzterWert;
               if funktionGeklickt then  z.zahlen:= funktional(eingabeZahl)
                                    else z.zahlen:= strToFloat(eingabeZahl);
               z.knuepfen:=symbol;
               letzterWert:=z;
               verknuepfung:=true;
             end
             else
               letzterWert.knuepfen:=symbol;  //die letzte Verknpfung kann
                                              //berschrieben werden
end;

//----------------------------------------------------------
//-----------Operatorkette abarbeiten-----------------------
//----------------------------------------------------------

procedure TForm1.gleichGeklickt(Sender: TObject);
var zwischenErgebnis : real;  n : integer;
begin
  zwischenErgebnis:=0.0;    // Delpi mosert sonst dauernd: keine Initialisierung
  if gleichWarGeklickt then exit;       //zweites '=' soll keine neue Berechnung
                                        //in der Kette auslsen
  if length(eingabeZahl)=0 then exit;   //'=' ohne vorhandene Operatoren

  if bildetKette then
//...... letzten Wert vor '=' schreiben .......................................
       begin
         new(z);
         z.vorgaenger:=letzterWert;
         if funktionGeklickt then z.zahlen:= funktional(eingabeZahl)
                             else z.zahlen:= strToFloat(eingabeZahl);
         z.knuepfen[1]:='=';
         letzterWert:=z;
         verknuepfung:=true;

//.......Schritt 1 zur arithmetischen Summe: beseitigen von Potenzen ...........

         while (z.vorgaenger <> nil)  do
           begin
             v:=z;
             z:=z.vorgaenger;
             if z.knuepfen[1]='e' then
               begin
                 if z.zahlen<0.0000000001 then begin errorMeldung ;exit;end
                  else zwischenergebnis:= exp(ln(z.zahlen)*v.zahlen);
                 z.zahlen:=zwischenergebnis;
                 z.knuepfen[1]:='*';
                 v.zahlen:=1.0;
               end;
           end;      

//.......Schritt 2 zur arithmetischen Summe: beseitigen von '/' ................

          z:=letzterWert;
          while (z.vorgaenger <> nil)  do
           begin
             v:=z;
             z:=z.vorgaenger;
             if z.knuepfen[1]='/' then
               begin
                 if abs(v.zahlen)<0.0000000001 then begin errorMeldung ;exit;end
                      else zwischenergebnis:= 1.0/v.zahlen;
                 v.zahlen:=zwischenergebnis;
                 z.knuepfen[1]:='*';
               end;
           end;

//.......Schritt 3 zur arithmetischen Summe: beseitigen von Prozent und ........
//.......Promille, falls sie wider Erwarten in einer Operatorkette .............
//.......auftauchen sollten  ...................................................

         z:=letzterWert;
         while (z.vorgaenger <> nil)  do
           begin
             v:=z;
             z:=z.vorgaenger;
             if z.knuepfen[1] in ['p','m'] then
               begin
                case z.knuepfen[1] of
                'p':  begin
                         if abs(v.zahlen)<0.000000001 then
                                    begin errorMeldung ;exit;end;
                         zwischenergebnis:= z.zahlen/v.zahlen*100.0;
                      end;
                'm':  begin
                         if abs(v.zahlen)<0.000000001 then
                                    begin errorMeldung ;exit;end;
                         zwischenergebnis:= z.zahlen/v.zahlen*1000.0;
                       end;
                end;
                z.zahlen:=zwischenergebnis;
                z.knuepfen[1]:='*';
                v.zahlen:=1.0;
               end;
            end;

//.............. Multiplikation ................................................

         z:=letzterWert;
         while (z.vorgaenger <> nil)  do
           begin
             v:=z;
             z:=z.vorgaenger;
             if z.knuepfen[1] ='*' then
               begin
                 zwischenergebnis:= z.zahlen*v.zahlen;
                 z.zahlen:=zwischenergebnis;
                 z.knuepfen:='+';
                 v.zahlen:=0.0;
               end;
           end;



    end //von bildetKette ; groe Eingangsschleife
    else

//..... noch keine Kette gebildet ( z.B. nach Start Eingabe 'sin(2)=' ) ........

        begin
          new(letzterWert);
          letzterWert.vorgaenger:=nil;
          if funktionGeklickt then letzterWert.zahlen:= funktional(eingabeZahl)
                              else letzterWert.zahlen:= strToFloat(eingabeZahl);
          letzterWert.knuepfen[1]:='=';
          verknuepfung:=True;
          bildetKette:=true;
        end;


//...Schritt 4 zur arithmetischen Summe: beseitigen von Minus-Verknpfungen ...

     z:=letzterWert;
     while z.vorgaenger <> nil do
        begin
          v:=z;
          z:=z.vorgaenger;
          if z.knuepfen[1]='-' then
                  begin
                    v.zahlen:=v.zahlen*(-1);
                    z.knuepfen[1]:='+';
                  end;

        end;


  //....... arithmetische Summe berechnen ......................................

    z:=letzterWert;
    zwischenErgebnis:=z.zahlen;
    while (z.vorgaenger <> nil)  do
     begin
       z:=z.vorgaenger;
       zwischenErgebnis:=zwischenErgebnis+z.zahlen ;
     end;



    clearAllGeklickt(self);           // Freigabe Listenkopf unter anderem
    if zwischenergebnis < 0
                  then eingabeFeld._inhalt:=floatToStr(zwischenErgebnis)
                   else   eingabeFeld._inhalt:='+'+floatToStr(zwischenErgebnis);


    eingabeZahl:= eingabeFeld._inhalt; //so kann mit dem Zwischenergebnis
                                       //als Startwert weitergerechnet werden ;
                                       //in clearAllGeklickt wird Stapel freige-
                                       //geben: eine neue Kette wird gebildet

    n:=1;
    dezimalPunkt:=false;               //Ist Ergebnis Real oder Integer ?
    while n<length(eingabeZahl)
    do begin
         n:=n+1;
         if eingabeZahl[n]='.' then dezimalPunkt:=true;
        end;

end;
//--------------Funktionsaufrufe--------------------------------
function tForm1.funktional(argument:string):real;
var lager : real;
begin
   lager:=0.0;                  //Delphi mosert sonst: lager nicht initialisiert
   case funktionsArt of
   's' : lager:=sin(strToFloat(argument));                            //Sinus
   'c' : lager:=cos(strToFloat(argument));                            //Kosinus
   't' : begin                                                        //Tangens
           if abs(cos(strToFloat(argument)))< 0.0000000001 then
                                   begin errorMeldung ;exit;end;
           lager:=tan(strToFloat(argument));
         end;
   'w' : begin                                                        //Wurzel
           if strToFloat(argument)< 0 then begin errorMeldung ;exit;end;
     //Wurzeln aus negativen Zahlen quittiert Delphi hier mit Ergebnis 0 !!??!
     //Fr eine Erklrung wre ich dankbar .....
           lager:=sqrt(strToFloat(argument));
         end;
   end;

   funktionGeklickt:=false;
   result:=lager;
   end;
//----------------------------------------------------------

end.
